#!/usr/bin/python

# Copyright 2012, Gurobi Optimization, Inc.

# Assign workers to shifts; each worker may or may not be available on a
# particular day. If the problem cannot be solved, add slack variables
# to determine which constraints cannot be satisfied, and how much
# they need to be relaxed.

from gurobipy import *

# Number of workers required for each shift
shifts, shiftRequirements = multidict({
  "Mon1":  3,
  "Tue2":  2,
  "Wed3":  4,
  "Thu4":  4,
  "Fri5":  5,
  "Sat6":  6,
  "Sun7":  5,
  "Mon8":  2,
  "Tue9":  2,
  "Wed10": 3,
  "Thu11": 4,
  "Fri12": 6,
  "Sat13": 7,
  "Sun14": 5 })

# Amount each worker is paid to work one shift
workers, pay = multidict({
  "Amy":   10,
  "Bob":   12,
  "Cathy": 10,
  "Dan":   8,
  "Ed":    8,
  "Fred":  9,
  "Gu":    11 })

# Worker availability
availability = tuplelist([
('Amy', 'Tue2'), ('Amy', 'Wed3'), ('Amy', 'Fri5'), ('Amy', 'Sun7'),
('Amy', 'Tue9'), ('Amy', 'Wed10'), ('Amy', 'Thu11'), ('Amy', 'Fri12'),
('Amy', 'Sat13'), ('Amy', 'Sun14'), ('Bob', 'Mon1'), ('Bob', 'Tue2'),
('Bob', 'Fri5'), ('Bob', 'Sat6'), ('Bob', 'Mon8'), ('Bob', 'Thu11'),
('Bob', 'Sat13'), ('Cathy', 'Wed3'), ('Cathy', 'Thu4'), ('Cathy', 'Fri5'),
('Cathy', 'Sun7'), ('Cathy', 'Mon8'), ('Cathy', 'Tue9'), ('Cathy', 'Wed10'),
('Cathy', 'Thu11'), ('Cathy', 'Fri12'), ('Cathy', 'Sat13'),
('Cathy', 'Sun14'), ('Dan', 'Tue2'), ('Dan', 'Wed3'), ('Dan', 'Fri5'),
('Dan', 'Sat6'), ('Dan', 'Mon8'), ('Dan', 'Tue9'), ('Dan', 'Wed10'),
('Dan', 'Thu11'), ('Dan', 'Fri12'), ('Dan', 'Sat13'), ('Dan', 'Sun14'),
('Ed', 'Mon1'), ('Ed', 'Tue2'), ('Ed', 'Wed3'), ('Ed', 'Thu4'),
('Ed', 'Fri5'), ('Ed', 'Sun7'), ('Ed', 'Mon8'), ('Ed', 'Tue9'),
('Ed', 'Thu11'), ('Ed', 'Sat13'), ('Ed', 'Sun14'), ('Fred', 'Mon1'),
('Fred', 'Tue2'), ('Fred', 'Wed3'), ('Fred', 'Sat6'), ('Fred', 'Mon8'),
('Fred', 'Tue9'), ('Fred', 'Fri12'), ('Fred', 'Sat13'), ('Fred', 'Sun14'),
('Gu', 'Mon1'), ('Gu', 'Tue2'), ('Gu', 'Wed3'), ('Gu', 'Fri5'),
('Gu', 'Sat6'), ('Gu', 'Sun7'), ('Gu', 'Mon8'), ('Gu', 'Tue9'),
('Gu', 'Wed10'), ('Gu', 'Thu11'), ('Gu', 'Fri12'), ('Gu', 'Sat13'),
('Gu', 'Sun14')
])

# Model
m = Model("assignment")

# Assignment variables: x[w,s] == 1 if worker w is assigned to shift s.
# Since an assignment model always produces integer solutions, we use
# continuous variables and solve as an LP.
x = {}
for w,s in availability:
    x[w,s] = m.addVar(ub=1, obj=pay[w], name=w+"."+s)

# The objective is to minimize the total pay costs
m.modelSense = GRB.MINIMIZE

# Update model to integrate new variables
m.update()

# Constraint: assign exactly shiftRequirements[s] workers to each shift s
reqCts = {}
for s in shifts:
    reqCts[s] = m.addConstr(
      quicksum(x[w,s] for w,s in availability.select('*', s)) ==
        shiftRequirements[s], s)

# Optimize
m.optimize()
status = m.status
if status == GRB.status.UNBOUNDED:
    print 'The model cannot be solved because it is unbounded'
    exit(0)
if status == GRB.status.OPTIMAL:
    print 'The optimal objective is', m.objVal
    exit(0)
if status != GRB.status.INF_OR_UNBD and status != GRB.status.INFEASIBLE:
    print 'Optimization was stopped with status', status
    exit(0)


# Add slack variables to make the model feasible
print 'The model is infeasible; adding slack variables'

# Set original objective coefficients to zero
m.setObjective(0.0)

# Add a new slack variable to each shift constraint so that the shifts can
# be satisfied
slacks = {}
for s,c in reqCts.items():
    slacks[s] = m.addVar(obj=1.0, name=s+"Slack", column=Column([1], [c]))

# Solve the model with slacks
m.optimize()
status = m.status
if status == GRB.status.INF_OR_UNBD or status == GRB.status.INFEASIBLE \
   or status == GRB.status.UNBOUNDED:
    print 'The model with slacks cannot be solved ' \
          'because it is infeasible or unbounded'
    exit(1)

if status != GRB.status.OPTIMAL:
    print 'Optimization was stopped with status', status
    exit(1)

print '\nSlack values:'
for sv in slacks.values():
    if sv.x > 1e-6:
        print sv.varName, '=', sv.x
